@@ -1,5 +1,5 @@
 # testing/assertions.py
-# Copyright (C) 2005-2022 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2024 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -7,7 +7,9 @@
 
 from __future__ import absolute_import
 
+from collections import defaultdict
 import contextlib
+from copy import copy
 import re
 import sys
 import warnings
@@ -143,14 +145,16 @@ def _expect_warnings(
     exc_cls,
     messages,
     regex=True,
+    search_msg=False,
     assert_=True,
     py2konly=False,
     raise_on_any_unexpected=False,
+    squelch_other_warnings=False,
 ):
 
     global _FILTERS, _SEEN, _EXC_CLS
 
-    if regex:
+    if regex or search_msg:
         filters = [re.compile(msg, re.I | re.S) for msg in messages]
     else:
         filters = list(messages)
@@ -188,19 +192,25 @@ def _expect_warnings(
                 exception = None
 
             if not exception or not issubclass(exception, _EXC_CLS):
-                return real_warn(msg, *arg, **kw)
+                if not squelch_other_warnings:
+                    return real_warn(msg, *arg, **kw)
+                else:
+                    return
 
             if not filters and not raise_on_any_unexpected:
                 return
 
             for filter_ in filters:
-                if (regex and filter_.match(msg)) or (
-                    not regex and filter_ == msg
+                if (
+                    (search_msg and filter_.search(msg))
+                    or (regex and filter_.match(msg))
+                    or (not regex and filter_ == msg)
                 ):
                     seen.discard(filter_)
                     break
             else:
-                real_warn(msg, *arg, **kw)
+                if not squelch_other_warnings:
+                    real_warn(msg, *arg, **kw)
 
         with mock.patch("warnings.warn", our_warn), mock.patch(
             "sqlalchemy.util.SQLALCHEMY_WARN_20", True
@@ -235,6 +245,17 @@ def _assert_no_stray_pool_connections():
     engines.testing_reaper.assert_all_closed()
 
 
+def int_within_variance(expected, received, variance):
+    deviance = int(expected * variance)
+    assert (
+        abs(received - expected) < deviance
+    ), "Given int value %s is not within %d%% of expected value %s" % (
+        received,
+        variance * 100,
+        expected,
+    )
+
+
 def eq_regex(a, b, msg=None):
     assert re.match(b, a), msg or "%r !~ %r" % (a, b)
 
@@ -313,8 +334,10 @@ def startswith_(a, fragment, msg=None):
 def eq_ignore_whitespace(a, b, msg=None):
     a = re.sub(r"^\s+?|\n", "", a)
     a = re.sub(r" {2,}", " ", a)
+    a = re.sub(r"\t", "", a)
     b = re.sub(r"^\s+?|\n", "", b)
     b = re.sub(r" {2,}", " ", b)
+    b = re.sub(r"\t", "", b)
 
     assert a == b, msg or "%r != %r" % (a, b)
 
@@ -357,6 +380,40 @@ def assert_raises_message(except_cls, msg, callable_, *args, **kwargs):
     )
 
 
+def assert_warns(except_cls, callable_, *args, **kwargs):
+    """legacy adapter function for functions that were previously using
+    assert_raises with SAWarning or similar.
+
+    has some workarounds to accommodate the fact that the callable completes
+    with this approach rather than stopping at the exception raise.
+
+
+    """
+    with _expect_warnings(except_cls, [".*"], squelch_other_warnings=True):
+        return callable_(*args, **kwargs)
+
+
+def assert_warns_message(except_cls, msg, callable_, *args, **kwargs):
+    """legacy adapter function for functions that were previously using
+    assert_raises with SAWarning or similar.
+
+    has some workarounds to accommodate the fact that the callable completes
+    with this approach rather than stopping at the exception raise.
+
+    Also uses regex.search() to match the given message to the error string
+    rather than regex.match().
+
+    """
+    with _expect_warnings(
+        except_cls,
+        [msg],
+        search_msg=True,
+        regex=False,
+        squelch_other_warnings=True,
+    ):
+        return callable_(*args, **kwargs)
+
+
 def assert_raises_message_context_ok(
     except_cls, msg, callable_, *args, **kwargs
 ):
@@ -378,6 +435,15 @@ class _ErrorContainer(object):
 
 @contextlib.contextmanager
 def _expect_raises(except_cls, msg=None, check_context=False):
+    if (
+        isinstance(except_cls, type)
+        and issubclass(except_cls, Warning)
+        or isinstance(except_cls, Warning)
+    ):
+        raise TypeError(
+            "Use expect_warnings for warnings, not "
+            "expect_raises / assert_raises"
+        )
     ec = _ErrorContainer()
     if check_context:
         are_we_already_in_a_traceback = sys.exc_info()[0]
@@ -437,6 +503,7 @@ class AssertsCompiledSQL(object):
         render_schema_translate=False,
         default_schema_name=None,
         from_linting=False,
+        check_param_order=True,
     ):
         if use_default_dialect:
             dialect = default.DefaultDialect()
@@ -450,8 +517,11 @@ class AssertsCompiledSQL(object):
 
             if dialect is None:
                 dialect = config.db.dialect
-            elif dialect == "default":
-                dialect = default.DefaultDialect()
+            elif dialect == "default" or dialect == "default_qmark":
+                if dialect == "default":
+                    dialect = default.DefaultDialect()
+                else:
+                    dialect = default.DefaultDialect(paramstyle="qmark")
                 dialect.supports_default_values = supports_default_values
                 dialect.supports_default_metavalue = supports_default_metavalue
             elif dialect == "default_enhanced":
@@ -583,7 +653,7 @@ class AssertsCompiledSQL(object):
         if checkparams is not None:
             eq_(c.construct_params(params), checkparams)
         if checkpositional is not None:
-            p = c.construct_params(params)
+            p = c.construct_params(params, escape_names=False)
             eq_(tuple([p[x] for x in c.positiontup]), checkpositional)
         if check_prefetch is not None:
             eq_(c.prefetch, check_prefetch)
@@ -603,6 +673,58 @@ class AssertsCompiledSQL(object):
                 },
                 check_post_param,
             )
+        if check_param_order and getattr(c, "params", None):
+
+            def get_dialect(paramstyle, positional):
+                cp = copy(dialect)
+                cp.paramstyle = paramstyle
+                cp.positional = positional
+                return cp
+
+            pyformat_dialect = get_dialect("pyformat", False)
+            pyformat_c = clause.compile(dialect=pyformat_dialect, **kw)
+            stmt = re.sub(r"[\n\t]", "", pyformat_c.string)
+
+            qmark_dialect = get_dialect("qmark", True)
+            qmark_c = clause.compile(dialect=qmark_dialect, **kw)
+            values = list(qmark_c.positiontup)
+            escaped = qmark_c.escaped_bind_names
+
+            for post_param in (
+                qmark_c.post_compile_params | qmark_c.literal_execute_params
+            ):
+                name = qmark_c.bind_names[post_param]
+                if name in values:
+                    values = [v for v in values if v != name]
+            positions = []
+            pos_by_value = defaultdict(list)
+            for v in values:
+                try:
+                    if v in pos_by_value:
+                        start = pos_by_value[v][-1]
+                    else:
+                        start = 0
+                    esc = escaped.get(v, v)
+                    pos = stmt.index("%%(%s)s" % (esc,), start) + 2
+                    positions.append(pos)
+                    pos_by_value[v].append(pos)
+                except ValueError:
+                    msg = "Expected to find bindparam %r in %r" % (v, stmt)
+                    assert False, msg
+
+            ordered = all(
+                positions[i - 1] < positions[i]
+                for i in range(1, len(positions))
+            )
+
+            expected = [v for _, v in sorted(zip(positions, values))]
+
+            msg = (
+                "Order of parameters %s does not match the order "
+                "in the statement %s. Statement %r" % (values, expected, stmt)
+            )
+
+            is_true(ordered, msg)
 
 
 class ComparesTables(object):
